% Illustration of EM for linear system identification
%
% Thomas Schn
% 22/4 - 2009

addpath('H:\Research\TechReports\2915EMExpl\FirstExMatlab\UNIT');    % Add path to UNIT toolbox

opt.miter = 100;    % Max number of EM iterations
opt.LLdec = 1e-6;   % Min decrease of log-likelihood
theta0 = 0.1;       % Initial guess for the parameter

%==========================================================
%===   Simulate the one dimensional state-space model   ===
%==========================================================
N = 500;          % Number of data
m.ss.A = 0.9;     m.ss.B = 0;
m.ss.C = 0.5;     m.ss.D = 0;
m.ss.Q = 0.1;     m.ss.S = 0;
m.ss.R = 0.1;

m.ss.X1 = 0;      m.ss.P1 = 0;  % Initial state (fully known)
x = zeros(1,N+1); y = zeros(1,N);
x(1) = m.ss.X1;
v = sqrt(m.ss.Q)*randn(1,N);    % Process noise sequence
e = sqrt(m.ss.R)*randn(1,N);    % Measurement noise sequence
for t=1:N
  x(t+1) = m.ss.A*x(t) + v(t);
  y(t)   = m.ss.C*x(t) + e(t);
end
z.y = y;
mEst = m;
mEst.ss.A = theta0;     % Initial parameter guess
%========================
%===   EM algorithm   ===
%========================
for k = 1:opt.miter
  % E step
  g     = ks(z,mEst,opt);     % Smoother
  LL(k) = -0.5*g.LL;          % Store the log-likelihood
  phi   = 0;   psi = 0;   sigma = 0;
  for t = 1:N
    phi   = phi + g.ss.xs(t+1)*g.ss.xs(t+1) + g.ss.Ps(1,1,t+1)^2;
    sigma = sigma + g.ss.xs(t).^2 + g.ss.Ps(1,1,t)^2;
    psi   = psi + g.ss.xs(t)*g.ss.xs(t+1) + g.ss.Ms(1,1,t);
  end;
  % M step
  a(k)      = psi/sigma;     % Store result
  mEst.ss.A = a(k);          % Update model

  % Compute the Q-function for the current estimate
  alpha = phi;
  beta  = 0;
  for t=1:N
    beta = beta + z.y(t)*g.ss.xs(t);
  end;
  aNow = [0:0.01:0.8 0.8:0.001:1 1.01:0.01:1.3];
  for i = 1:length(aNow)
    Q1      = -N*log(0.2*pi) - 5*sum(z.y.^2) - 5*alpha;
    Q2      = 5*beta + 10*psi*aNow(i) - 5*sigma*(aNow(i)^2+1/4);
    Q{k}(i) = Q1 + Q2;
  end;

  if k>1          % Check termination condition
    if LL(k)-LL(k-1) < opt.LLdec
      break;
    end;
  end;
end;
a = [theta0 a];

% Compute the log-likelihood function
for i = 1:length(aNow)
  m.ss.A = aNow(i);
  g      = kf(z,m,opt);      % Kalman filter
  Ltmp   = -(N/2)*log(2*pi);
  for t = 2:N
    CovInn   = g.ss.Ri(1,1,t)^2;
    Lpart1   = -(1/2)*((z.y(t) - m.ss.C*g.ss.xp(t))^2)/CovInn;
    Lpart2   = -(1/2)*log(det(CovInn));
    Ltmp     = Ltmp + Lpart1 + Lpart2;
  end;
  LL(i) = Ltmp;
end;